@@ -1,4 +1,7 @@
-from itertools import product
+"""A collection of functions for manipulating data in CasADi, in particular, on how
+to convert to and from numpy arrays and CasADi symbolic variables, and how to find
+the indices of a symbolic variable in a vector of symbols."""
+
 from typing import Union
 
 import casadi as cs
@@ -7,25 +10,28 @@ import numpy.typing as npt
 
 
 def array2cs(x: np.ndarray) -> Union[cs.SX, cs.MX]:
-    """Converts numpy array `x` of scalar symbolic variable to a single symbolic
-    instance. Opposite to `array2cs`. Note that all entries in `x` must have the same
-    type, either SX or MX.
+    """Converts the numpy array ``x`` containing scalar symbolic variables to a single
+    symbolic variable (see :func:`cs2array` for the opposite functionality).
 
     Parameters
     ----------
-    x : np.ndarray
-        Array whose entries are either MX or SX. In case `x` is SX, MX or DM, it is
-        returned immediately.
+    x : numpy array of casadi.SX or MX
+        Array whose entries are either MX or SX.
 
     Returns
     -------
     casadi.SX or MX
-        A single SX or MX instance whose entries are `x`'s entries.
+        A single SX or MX instance whose entries are the ``x``'s entries.
 
     Raises
     ------
     ValueError
-        Raises if the array is empty (zero dimensions), or if it has more than 2 dims.
+        Raises if the array is empty (zero dimensions), or if it has more than 2
+        dimensions.
+
+    Notes
+    -----
+    Note that all entries in ``x`` must have the same type, either SX or MX.
     """
     if isinstance(x, (cs.SX, cs.MX, cs.DM)):
         return x
@@ -35,14 +41,13 @@ def array2cs(x: np.ndarray) -> Union[cs.SX, cs.MX]:
     ndim = x.ndim
     if ndim == 0:
         return x.item()
-    elif ndim == 1:
-        indices = range(x.shape[0])
-        shape = (x.shape[0], 1)
+    if ndim == 1:
+        numel = x.size
+        shape = (numel, 1)
+        indices = range(numel)
     elif ndim == 2:
-        indices = product(  # type: ignore[assignment]
-            range(x.shape[0]), range(x.shape[1])
-        )
-        shape = x.shape  # type: ignore[assignment]
+        shape = x.shape
+        indices = np.ndindex(shape)
     else:
         raise ValueError("Can only convert 1D and 2D arrays to CasADi SX or MX.")
     m: Union[cs.SX, cs.MX] = cls(*shape)
@@ -51,12 +56,12 @@ def array2cs(x: np.ndarray) -> Union[cs.SX, cs.MX]:
     return m
 
 
-def cs2array(x: Union[cs.MX, cs.SX]) -> npt.NDArray[np.object_]:
-    """Converts casadi symbolic variable `x` to a numpy array of scalars. Opposite to
-    `array2cs`.
+def cs2array(x: Union[cs.MX, cs.SX]) -> np.ndarray:
+    """Converts the symbolic variable ``x`` to a numpy array of scalar symbolic
+    variables (see :func:`array2cs` for the opposite functionality).
 
-    Inspired by
-    https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/tools.py
+    Inspired by the implementation in
+    `MPCTools <https://bitbucket.org/rawlings-group/mpc-tools-casadi/src/master/mpctools/tools.py>`_.
 
     Parameters
     ----------
@@ -65,7 +70,7 @@ def cs2array(x: Union[cs.MX, cs.SX]) -> npt.NDArray[np.object_]:
 
     Returns
     -------
-    np array of cs.SX or MX
+    numpy array of casadi.SX or MX
         The array containing the symbolic variable scalars.
     """
     if isinstance(x, np.ndarray):
@@ -75,8 +80,7 @@ def cs2array(x: Union[cs.MX, cs.SX]) -> npt.NDArray[np.object_]:
 
     shape = x.shape
     y = np.empty(shape, object)
-    indices = product(range(shape[0]), range(shape[1]))
-    for idx in indices:
+    for idx in np.ndindex(shape):
         y[idx] = x[idx]
     return y
 
@@ -84,7 +88,7 @@ def cs2array(x: Union[cs.MX, cs.SX]) -> npt.NDArray[np.object_]:
 def find_index_in_vector(
     V: Union[cs.SX, cs.MX], a: Union[cs.SX, cs.MX]
 ) -> npt.NDArray[np.int_]:
-    """Finds the indices of `a` in `V`.
+    """Finds the indices of ``a`` in ``V``.
 
     Parameters
     ----------
@@ -95,13 +99,13 @@ def find_index_in_vector(
 
     Returns
     -------
-    array of int
-        The indices of `a` in `V`.
+    array of integers
+        The indices of ``a`` in ``V``.
 
     Raises
     ------
     ValueError
-        Raises if `V` or `a` are not vectors.
+        Raises if ``V`` or ``a`` are not vectors.
     """
     if not V.is_vector() or not a.is_vector():
         raise ValueError("`V` and `a` must be vectors.")
