@@ -1,4 +1,3 @@
-import warnings
 from collections.abc import Iterator, Sequence
 from itertools import count
 from typing import Any, ClassVar, Literal, Optional, TypeVar, Union
@@ -9,24 +8,50 @@ import numpy.typing as npt
 from joblib import Memory
 
 from ..core.debug import NlpDebug
-from ..util.io import SupportsDeepcopyAndPickle
 from .objective import HasObjective
 
 SymType = TypeVar("SymType", cs.SX, cs.MX)
 
 
-class Nlp(HasObjective[SymType], SupportsDeepcopyAndPickle):
-    """
-    The generic NLP class is a controller that solves a (possibly, nonlinear)
-    optimization problem to yield a (possibly, sub-) optimal solution.
+class Nlp(HasObjective[SymType]):
+    r"""The generic NLP class is a controller that solves a (possibly, nonlinear)
+    optimization problem to yield a (possibly, sub-) optimal solution. This is a generic
+    implementation in the sense that it does not solve a particular problem, but only
+    offers the generic methods to build one (e.g., variables, constraints, objective,
+    solver).
 
-    This is a generic class in the sense that it does not solve a particular
-    problem, but only offers the generic methods to build one (e.g., variables,
-    constraints, objective, solver).
+    Parameters
+    ----------
+    sym_type : {"SX", "MX"}, optional
+        The CasADi symbolic variable type to use in the NLP, by default ``"SX"``.
+    remove_redundant_x_bounds : bool, optional
+        If ``True``, then redundant entries in :meth:`lbx` and :meth:`ubx` are removed
+        when properties :meth:`h_lbx` and :meth:`h_ubx` are called. See these two
+        properties for more details. By default, ``True``.
+    cache : joblib.Memory, optional
+        Optional cache to avoid computing the same exact NLP more than once. By default,
+        no caching occurs.
+    name : str, optional
+        Name of the NLP scheme. If `None`, it is automatically assigned.
+    debug : bool, optional
+        If ``True``, the NLP logs in the :meth:`debug` property information regarding
+        the creation of parameters, variables and constraints. By default, ``False``.
+
+    Raises
+    ------
+    AttributeError
+        Raises if the specified CasADi's symbolic type is neither ``"SX"`` nor ``"MX"``.
+
+    Notes
+    -----
+    Constraints are internally handled in their canonical form, i.e., :math:`g(x,p) = 0`
+    and :math:`h(x,p) \leq 0`. The objective :math:`f(x,p)` is always a scalar function
+    to be minimized.
     """
 
     __ids: ClassVar[Iterator[int]] = count(0)
     is_multi: ClassVar[bool] = False
+    """Flag to indicate that this is not a multistart NLP."""
 
     def __init__(
         self,
@@ -36,34 +61,9 @@ class Nlp(HasObjective[SymType], SupportsDeepcopyAndPickle):
         name: Optional[str] = None,
         debug: bool = False,
     ) -> None:
-        """Creates an NLP instance.
-
-        Parameters
-        ----------
-        sym_type : "SX" or "MX", optional
-            The CasADi symbolic variable type to use in the NLP, by default "SX".
-        remove_redundant_x_bounds : bool, optional
-            If `True`, then redundant entries in `lbx` and `ubx` are removed when
-            properties `h_lbx` and `h_ubx` are called. See these two properties for more
-            details. By default, `True`.
-        cache : joblib.Memory, optional
-            Optional cache to avoid computing the same exact NLP more than once. By
-            default, no caching occurs.
-        name : str, optional
-            Name of the NLP scheme. If `None`, it is automatically assigned.
-        debug : bool, optional
-            If `True`, the NLP logs in the `debug` property information regarding the
-            creation of parameters, variables and constraints. By default, `False`.
-
-        Raises
-        ------
-        AttributeError
-            Raises if the specified CasADi's symbolic type is neither "SX" nor "MX".
-        """
         id = next(self.__ids)
         name = f"{self.__class__.__name__}{id}" if name is None else name
-        HasObjective.__init__(self, sym_type, remove_redundant_x_bounds, cache, name)
-        SupportsDeepcopyAndPickle.__init__(self)
+        super().__init__(sym_type, remove_redundant_x_bounds, cache, name)
         self.id = id
         self._debug = NlpDebug() if debug else None
 
@@ -82,7 +82,7 @@ class Nlp(HasObjective[SymType], SupportsDeepcopyAndPickle):
         """Returns the original NLP of the wrapper."""
         return self
 
-    def is_wrapped(self, *args: Any, **kwargs: Any) -> bool:
+    def is_wrapped(self, *_: Any, **__: Any) -> bool:
         """Gets whether the NLP instance is wrapped or not by the given wrapper type."""
         return False
 
@@ -92,14 +92,15 @@ class Nlp(HasObjective[SymType], SupportsDeepcopyAndPickle):
             self._debug.register("p", name, shape)
         return out
 
-    def variable(  # type: ignore[override]
+    def variable(
         self,
         name: str,
         shape: tuple[int, int] = (1, 1),
+        discrete: bool = False,
         lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
         ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
     ) -> tuple[SymType, SymType, SymType]:
-        out = super().variable(name, shape, lb, ub)
+        out = super().variable(name, shape, discrete, lb, ub)
         if self._debug is not None:
             self._debug.register("x", name, shape)
         return out
@@ -126,67 +127,66 @@ class Nlp(HasObjective[SymType], SupportsDeepcopyAndPickle):
         name_in: Optional[Sequence[str]] = None,
         name_out: Optional[Sequence[str]] = None,
         opts: Optional[dict[Any, Any]] = None,
+        mx_prewrap: bool = False,
     ) -> cs.Function:
-        """Converts the optimization problem to an MX symbolic function. If the
-        NLP is modelled in SX, the function will still be converted in MX since
-        the IPOPT interface cannot expand SX for now.
+        """Converts the optimization problem to an ``SX`` or ``MX`` symbolic
+        :class:`casadi.Function`. If the NLP is modelled in ``SX``, the function can be
+        pre-wrapped in ``MX``.
 
         Parameters
         ----------
         name : str
             Name of the function.
-        ins : Sequence of cs.SX or MX
-            Input variables of the function. These must be expressions
-            providing the parameters of the NLP and the initial conditions of
-            the primal variables `x`.
-        outs : Sequence of cs.SX or MX
-            Output variables of the function. These must be expressions
-            depending on the primal variable `x`, parameters `p`, and dual
-            variables `lam_g`, `lam_h`, `lam_lbx`, `lam_ubx` of the NLP.
-        name_in : Sequence of str, optional
-            Name of the inputs, by default None.
-        name_out : Sequence of str, optional
-            Name of the outpus, by default None.
-        opts : Dict[Any, Any], optional
-            Options to be passed to `casadi.Function`, by default None.
+        ins : sequence of casadi.SX or MX
+            Input variables of the function. These must be expressions providing the
+            parameters of the NLP and the initial conditions of the primal variables
+            ``x``.
+        outs : sequence of casadi.SX or MX
+            Output variables of the function. These must be expressions depending on the
+            primal variable ``x``, parameters ``p``, and dual variables ``lam_g``,
+            ``lam_h``, ``lam_lbx``, ``lam_ubx`` of the NLP.
+        name_in : sequence of str, optional
+            Name of the inputs, by default ``None``.
+        name_out : sequence of str, optional
+            Name of the outpus, by default ``None``.
+        opts : dict[Any, Any], optional
+            Options to be passed to :class:`casadi.Function`, by default ``None``.
+        mx_prewrap : bool, optional
+            If ``True``, wraps the CasADi interface in an ``MX`` wrapper prior to
+            turning it into the function. This is useful when the NLP is defined in
+            ``SX`` but the interface is only supported in ``MX``. By default, ``False``.
 
         Returns
         -------
         casadi.Function
-            The NLP solver as a `casadi.Function`.
+            The NLP solver as an instance of :class:`casadi.Function`.
 
         Raises
         ------
         RuntimeError
-            Raises if the solver is uninitialized; or if the input or output expressions
+            Raises if the solver is uninitialized, or if the input or output expressions
             have free variables that are not provided or cannot be computed by the
             solver.
         """
-        if self._sym_type is cs.SX:
-            warnings.warn(
-                "The IPOPT interface does not support SX expansion, "
-                "so the function must be wrapped in MX.",
-                RuntimeWarning,
-            )
         S = self._solver
         if S is None:
             raise RuntimeError("Solver not yet initialized.")
+        n_ins = len(ins)
+        n_outs = len(outs)
 
         # converts inputs/outputs to/from variables and parameters
-        n_outs = len(outs)
-        Fin = cs.Function("Fin", ins, [self._x, self._p])
+        Fin = cs.Function("Fin", ins, (self._x, self._p))
         Fout = cs.Function(
             "Fout",
-            [self._p, self._x, self._lam_g, self._lam_h, self._lam_lbx, self._lam_ubx],
+            (self._p, self._x, self._lam_g, self._lam_h, self._lam_lbx, self._lam_ubx),
             outs,
         )
 
         # call the solver
-        if self._sym_type is cs.SX:
+        if mx_prewrap:
             Fin = Fin.wrap()
             Fout = Fout.wrap()
-            ins = [Fin.mx_in(i) for i in range(len(ins))]
-            outs = [Fout.mx_out(i) for i in range(len(outs))]
+            ins = [Fin.mx_in(i) for i in range(n_ins)]
         x0, p = Fin(*ins)
         sol = S(
             x0=x0,
@@ -203,10 +203,9 @@ class Nlp(HasObjective[SymType], SupportsDeepcopyAndPickle):
         lam_h = sol["lam_g"][self.ng :, :]
         lam_lbx = -cs.fmin(sol["lam_x"], 0)[self.nonmasked_lbx_idx, :]
         lam_ubx = cs.fmax(sol["lam_x"], 0)[self.nonmasked_ubx_idx, :]
-        Fsol = cs.Function("Fsol", ins, [x, lam_g, lam_h, lam_lbx, lam_ubx])
 
         # build final function
-        final_outs = Fout(p, *Fsol(*ins))
+        final_outs = Fout(p, x, lam_g, lam_h, lam_lbx, lam_ubx)
         if n_outs == 1:
             final_outs = [final_outs]
         args = [name, ins, final_outs]
