@@ -1,28 +1,53 @@
-from itertools import chain
+from inspect import signature
 from typing import Callable, Literal, Optional, TypeVar, Union
 
 import casadi as cs
 import numpy as np
 import numpy.typing as npt
 
-from csnlp.multistart.multistart_nlp import _chained_subevalf, _n
+from csnlp.multistart.multistart_nlp import _chained_substitute, _n
 
 from ..wrapper import Nlp
-from .mpc import Mpc
+from .mpc import Mpc, _callable2csfunc, _create_ati_mats
 from .mpc import _n as _name_init_state
 
 SymType = TypeVar("SymType", cs.SX, cs.MX)
+MatType = TypeVar("MatType", SymType, cs.DM, np.ndarray)
 
 
 class ScenarioBasedMpc(Mpc[SymType]):
-    """Implementation of the Scenario-based Model Predictive Control [1], here referred
-    to as SCMPC, a well-known stochastic MPC formulation.
+    """Implementation of the Scenario-based Model Predictive Control
+    :cite:`schildbach_scenario_2014`, here referred to as SCMPC, a well-known stochastic
+    MPC formulation.
 
-    References
+    Parameters
     ----------
-    [1] Schildbach, G., Fagiano, L., Frei, C. and Morari, M., 2014. The scenario
-        approach for stochastic model predictive control with bounds on closed-loop
-        constraint violations. Automatica, 50(12), pp.3009-3018.
+    nlp : Nlp
+        NLP scheme to be wrapped
+    n_scenarios : int
+        Number of scenarios to be considered in the scenario-based MPC formulation. Must
+        be a positive integer.
+    prediction_horizon : int
+        A positive integer for the prediction horizon of the MPC controller.
+    control_horizon : int, optional
+        A positive integer for the control horizon of the MPC controller. If not given,
+        it is set equal to the control horizon.
+    input_spacing : int, optional
+        Spacing between independent input actions. This argument allows to reduce the
+        number of free actions along the control horizon by allowing only the first
+        action every ``n`` to be free, and the following ``n-1`` to be fixed equal to
+        that action (where ``n`` is given by ``input_spacing``). By default, no spacing
+        is allowed, i.e., ``1``.
+    shooting : 'single' or 'multi', optional
+        Type of approach in the direct shooting for parametrizing the control
+        trajectory. See Section 8.5 in :cite:`rawlings_model_2017`. By default, direct
+        shooting is used.
+
+    Raises
+    ------
+    ValueError
+        Raises if the shooting method is invalid; or if any of the horizons are invalid;
+        or if the number of scenarios is not a positive integer.
     """
 
     def __init__(
@@ -34,41 +59,6 @@ class ScenarioBasedMpc(Mpc[SymType]):
         input_spacing: int = 1,
         shooting: Literal["single", "multi"] = "multi",
     ) -> None:
-        """Initializes the Scenarion-based MPC (SCMPC) wrapper around the NLP instance.
-
-        Parameters
-        ----------
-        nlp : Nlp
-            NLP scheme to be wrapped
-        n_scenarios : int
-            Number of scenarios to be considered in the scenario-based MPC formulation.
-            Must be a positive integer.
-        prediction_horizon : int
-            A positive integer for the prediction horizon of the MPC controller.
-        control_horizon : int, optional
-            A positive integer for the control horizon of the MPC controller. If not
-            given, it is set equal to the control horizon.
-        input_spacing : int, optional
-            Spacing between independent input actions. This argument allows to reduce
-            the number of free actions along the control horizon by allowing only the
-            first action every `n` to be free, and the following `n-1` to be fixed equal
-            to that action (where `n` is given by `input_spacing`). By default, no
-            spacing is allowed, i.e., 1.
-        shooting : 'single' or 'multi', optional
-            Type of approach in the direct shooting for parametrizing the control
-            trajectory. See [1, Section 8.5]. By default, direct shooting is used.
-
-        Raises
-        ------
-        ValueError
-            Raises if the shooting method is invalid; or if any of the horizons are
-            invalid; or if the number of scenarios is not a positive integer.
-
-        References
-        ----------
-        [1] Rawlings, J.B., Mayne, D.Q. and Diehl, M., 2017. Model Predictive Control:
-            theory, computation, and design (Vol. 2). Madison, WI: Nob Hill Publishing.
-        """
         if n_scenarios < 1:
             raise ValueError("The number of scenarios must be a positive integer.")
         super().__init__(
@@ -120,6 +110,7 @@ class ScenarioBasedMpc(Mpc[SymType]):
         self,
         name: str,
         size: int = 1,
+        discrete: bool = False,
         lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
         ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
         bound_initial: bool = True,
@@ -134,17 +125,19 @@ class ScenarioBasedMpc(Mpc[SymType]):
             Name of the state.
         size : int
             Size of the state (assumed to be a vector).
+        discrete : bool, optional
+            Flag indicating if the state is discrete. Defaults to ``False``.
         lb : array_like, casadi.DM, optional
-            Hard lower bound of the state, by default -np.inf.
+            Hard lower bound of the state, by default ``-np.inf``.
         ub : array_like, casadi.DM, optional
-            Hard upper bound of the state, by default +np.inf.
+            Hard upper bound of the state, by default ``+np.inf``.
         bound_initial : bool, optional
-            If `False`, then the upper and lower bounds on the initial state are not
-            imposed, i.e., set to `+/- np.inf` (since the initial state is constrained
+            If ``False``, then the upper and lower bounds on the initial state are not
+            imposed, i.e., set to ``+/- np.inf`` (since the initial state is constrained
             to be equal to the current state of the system, it is sometimes advantageous
-            to remove its bounds). By default `True`.
+            to remove its bounds). By default ``True``.
         bound_terminal : bool, optional
-            Same as above, but for the terminal state. By default `True`.
+            Same as above, but for the terminal state. By default ``True``.
 
         Returns
         -------
@@ -169,14 +162,15 @@ class ScenarioBasedMpc(Mpc[SymType]):
             since these can only be set after the dynamics have been set via the
             `constraint` method.
         """
+        N = self._prediction_horizon
         if self._is_multishooting:
-            shape = (size, self._prediction_horizon + 1)
+            shape = (size, N + 1)
             lb = np.broadcast_to(lb, shape).astype(float)
             ub = np.broadcast_to(ub, shape).astype(float)
-            if bound_initial:
+            if not bound_initial:
                 lb[:, 0] = -np.inf
                 ub[:, 0] = +np.inf
-            if bound_terminal:
+            if not bound_terminal:
                 lb[:, -1] = -np.inf
                 ub[:, -1] = +np.inf
         elif np.any(lb != -np.inf) or np.any(ub != +np.inf):
@@ -184,8 +178,6 @@ class ScenarioBasedMpc(Mpc[SymType]):
                 "in single shooting, lower and upper state bounds can only be "
                 "created after the dynamics have been set"
             )
-        else:
-            shape = (size, 1)
 
         # create as many states as scenarions, but only one initial state
         x0_name = _name_init_state(name)
@@ -196,7 +188,7 @@ class ScenarioBasedMpc(Mpc[SymType]):
         for i in range(self._n_scenarios):
             name_i = _n(name, i)
             if self._is_multishooting:
-                x_i = self.nlp.variable(name_i, shape, lb, ub)[0]
+                x_i = self.nlp.variable(name_i, shape, discrete, lb, ub)[0]
                 self.nlp.constraint(_n(x0_name, i), x_i[:, 0], "==", x0)
             else:
                 x_i = None
@@ -204,9 +196,7 @@ class ScenarioBasedMpc(Mpc[SymType]):
             self._states[name_i] = x_i
 
         # create also a single symbol for the state
-        x_single = (
-            self.nlp._sym_type.sym(name, shape) if self._is_multishooting else None
-        )
+        x_single = self.nlp._sym_type.sym(name, size, N + 1)
         self.single_states[name] = x_single
         return x_single, xs, x0
 
@@ -219,13 +209,13 @@ class ScenarioBasedMpc(Mpc[SymType]):
         name : str
             Name of the disturbance.
         size : int, optional
-            Size of the disturbance (assumed to be a vector). Defaults to 1.
+            Size of the disturbance (assumed to be a vector). Defaults to ``1``.
 
         Returns
         -------
         single disturbance : casadi.SX or MX
             Symbol corresponding to the disturbance of a single scenario. See the note
-            for method `state`.
+            for :meth:``state``.
         disturbances : list of casadi.SX or MX
             The symbols for the new disturbances of each scenario in the SCMPC
             controller.
@@ -250,19 +240,20 @@ class ScenarioBasedMpc(Mpc[SymType]):
         soft: bool = False,
         simplify: bool = True,
     ) -> tuple[list[SymType], ...]:
-        """Similarly to `Nlp.constraint`, adds a constraint to the NLP scheme. However,
-        instead of manually creating the constraint for each scenario, this method
-        allows to define only one constraint expression for a single scenario, which is
-        then automatically declined for all scenarios. The symbolical expression must
-        be made up of the single scenario states and disturbances, returned as first
-        output by the methods `state` and `disturbance`, respectively.
+        """Similarly to :meth:`csnlp.wrappers.Mpc.constraint`, adds a constraint to the
+        MPC scheme. However, instead of manually creating the constraint for each
+        scenario, this method allows to define only one constraint expression for a
+        single scenario, which is then automatically declined for all scenarios. The
+        symbolical expression must be made up of the single scenario states and
+        disturbances, returned as first output by the methods :meth:`state` and
+        :meth:`disturbance`, respectively. Note that the return types are lists of
+        symbolical variables.
 
-        Note that the return types are list of symbolical variables.
         Returns
         -------
         exprs : list of casadi.SX or MX
-            The constraint expression in canonical form, i.e., `g(x,u) = 0` or
-            `h(x,u) <= 0`, for each scenario.
+            The constraint expression in canonical form, i.e., :math:`g(x,u) = 0` or
+            :math:`h(x,u) <= 0`, for each scenario.
         lams : list of casadi.SX or MX
             The symbol corresponding to the constraint's multipliers, for each scenario.
         single slack : casadi.SX or MX
@@ -283,15 +274,7 @@ class ScenarioBasedMpc(Mpc[SymType]):
         if soft:
             slacks = []
         for i in range(self._n_scenarios):
-            expr_i = _chained_subevalf(
-                expr,
-                self.single_states,
-                self.states_i(i),
-                self.single_disturbances,
-                self.disturbances_i(i),
-                eval=False,
-            )
-
+            expr_i = self._chained_substitution_for_scenario_i(expr, i)
             out = self.constraint(_n(name, i), expr_i, op, 0, soft)
             cons.append(out[0])
             lams.append(out[1])
@@ -306,82 +289,174 @@ class ScenarioBasedMpc(Mpc[SymType]):
         return cons, lams
 
     def minimize_from_single(self, objective: SymType) -> None:
-        """Similarly to `Nlp.minimize`, adds the objective to be minimized to the NLP
-        scheme. However, instead of manually creating the objective for each scenario,
-        this method allows to define only one expression for a single scenario, which is
-        then automatically declined and summed for all scenarios. The symbolical
-        expression must be made up of the single scenario states, disturbances, and
-        slacks, returned as first output by the methods `state`, `disturbance`, and
-        `constraint_from_single`, respectively.
+        """Similarly to :meth:`csnlp.Nlp.minimize`, adds the objective to be minimized
+        to the NLP scheme. However, instead of manually creating the objective for each
+        scenario, this method allows to define only one expression for a single
+        scenario, which is then automatically declined and summed for all scenarios. The
+        symbolical expression must be made up of the single scenario states,
+        disturbances, and slacks, returned as first output by the methods :meth:`state`,
+        :meth:`disturbance`, and :meth:`constraint_from_single`, respectively.
         """
-        objective = cs.simplify(cs.cse(objective)) / self._n_scenarios
         return self.nlp.minimize(
             sum(
-                _chained_subevalf(
-                    objective,
-                    self.single_states,
-                    self.states_i(i),
-                    self.single_disturbances,
-                    self.disturbances_i(i),
-                    self.single_slacks,
-                    self.slacks_i(i),
-                    eval=False,
-                )
+                self._chained_substitution_for_scenario_i(objective, i)
                 for i in range(self._n_scenarios)
             )
+            / self._n_scenarios
         )
 
-    def set_dynamics(
+    def set_affine_dynamics(
+        self,
+        A: MatType,
+        B: MatType,
+        D: MatType,
+        c: Optional[MatType] = None,
+        parallelization: Literal[
+            "serial", "unroll", "inline", "thread", "openmp"
+        ] = "thread",
+        max_num_threads: Optional[int] = None,
+    ) -> tuple[
+        Optional[MatType], Optional[MatType], Optional[MatType], Optional[MatType]
+    ]:
+        if D is None:
+            raise ValueError(
+                "The dynamics matrix D must be given, as SCMPC is a tool to account for"
+                " stochastic disturbances, and if there are none, a nominal MPC should "
+                "suffice (see `Mpc` wrapper)."
+            )
+        if max_num_threads is None:
+            max_num_threads = max(self._prediction_horizon, self._n_scenarios)
+        return super().set_affine_dynamics(A, B, D, c, parallelization, max_num_threads)
+
+    def set_nonlinear_dynamics(
         self,
         F: Union[
             cs.Function,
             Callable[[tuple[npt.ArrayLike, ...]], tuple[npt.ArrayLike, ...]],
         ],
-        n_in: Optional[int] = None,
-        n_out: Optional[int] = None,
+        parallelization: Literal[
+            "serial", "unroll", "inline", "thread", "openmp"
+        ] = "thread",
+        max_num_threads_or_unrolling_base: Optional[int] = None,
     ) -> None:
-        if isinstance(F, cs.Function):
-            n_in = F.n_in()
-        if n_in != 3 or self.nd == 0:
+        n_in = F.n_in() if isinstance(F, cs.Function) else len(signature(F).parameters)
+        nd = self.nd
+        if n_in != 3 or nd == 0:
             raise ValueError(
                 "The dynamics function must have 3 arguments: the state, the action, "
                 "and the disturbance. This is because SCMPC is a tool to account for "
                 "stochastic disturbances, and if there are none, a nominal MPC should "
                 "suffice (see `Mpc` wrapper)."
             )
-        return super().set_dynamics(F, n_in, n_out)
 
-    def _multishooting_dynamics(self, F: cs.Function, _: int, n_out: int) -> None:
+        if not isinstance(F, cs.Function):
+            F = _callable2csfunc(F, self.nlp.sym_type, (self.ns, self.na, nd))
+
+        if max_num_threads_or_unrolling_base is None:
+            max_num_threads_or_unrolling_base = max(
+                self._prediction_horizon, self._n_scenarios
+            )
+
+        return super().set_nonlinear_dynamics(
+            F, parallelization, max_num_threads_or_unrolling_base
+        )
+
+    def _set_singleshooting_affine_dynamics(
+        self, A: MatType, B: MatType, D: MatType, c: Optional[MatType]
+    ) -> tuple[MatType, MatType, Optional[MatType], Optional[MatType]]:
+        disturbance_names = self.single_disturbances.keys()
+        X0 = cs.vcat(self._initial_states.values())
+        U = cs.vec(cs.vcat(self._actions_exp.values()))  # NOTE: different from vvcat!
+        D_all = cs.hcat(
+            [
+                cs.vec(
+                    cs.vcat([self._disturbances[_n(n, i)] for n in disturbance_names])
+                )
+                for i in range(self._n_scenarios)
+            ]
+        )
+
+        F, G, H, L = _create_ati_mats(self._prediction_horizon, A, B, D, c)
+        X_next_pred = F @ X0 + G @ U + H @ D_all
+        if L is not None:
+            X_next_pred += L
+
+        state_names = self.single_states.keys()
+        N = self._prediction_horizon
+        ns = A.shape[0]
+        cumsizes = np.cumsum([0] + [s.shape[0] for s in self._initial_states.values()])
+        for i in range(self._n_scenarios):
+            X_i = cs.vertcat(X0, X_next_pred[:, i]).reshape((ns, N + 1))
+            X_i_split = cs.vertsplit(X_i, cumsizes)
+            for n, x in zip(state_names, X_i_split):
+                self._states[_n(n, i)] = x
+        return F, G, H, L
+
+    def _set_multishooting_nonlinear_dynamics(
+        self,
+        F: cs.Function,
+        _: int,
+        parallelization: Literal["serial", "unroll", "inline", "thread", "openmp"],
+        max_num_threads: int,
+    ) -> None:
         state_names = self.single_states.keys()
         disturbance_names = self.single_disturbances.keys()
         U = cs.vcat(self._actions_exp.values())
+        X_all_ = []
+        X_all_next_ = []
+        D_all_ = []
         for i in range(self._n_scenarios):
             X_i = cs.vcat([self._states[_n(n, i)] for n in state_names])
-            D_i = cs.vcat([self._disturbances[_n(n, i)] for n in disturbance_names])
-            xs_i_next = []
-            for k in range(self._prediction_horizon):
-                x_i_next = F(X_i[:, k], U[:, k], D_i[:, k])
-                if n_out != 1:
-                    x_i_next = x_i_next[0]
-                xs_i_next.append(x_i_next)
-            self.constraint(_n("dyn", i), cs.hcat(xs_i_next), "==", X_i[:, 1:])
+            X_all_.append(X_i[:, :-1])
+            X_all_next_.append(X_i[:, 1:])
+            D_all_.append(
+                cs.vcat([self._disturbances[_n(n, i)] for n in disturbance_names])
+            )
+        X_all = cs.hcat(X_all_)
+        X_all_next = cs.hcat(X_all_next_)
+        D_all = cs.hcat(D_all_)
 
-    def _singleshooting_dynamics(self, F: cs.Function, _: int, n_out: int) -> None:
+        Fmap = F.map(
+            self._prediction_horizon * self._n_scenarios,
+            parallelization,
+            max_num_threads,
+        )
+        X_all_next_pred = Fmap(X_all, U, D_all)
+        self.constraint("dyn", X_all_next, "==", X_all_next_pred)
+
+    def _set_singleshooting_nonlinear_dynamics(
+        self, F: cs.Function, _: int, base: int
+    ) -> None:
         disturbance_names = self.single_disturbances.keys()
-        Xk_shared = cs.vcat(self._initial_states.values())
-        cumsizes = np.cumsum([0] + [s.shape[0] for s in self._initial_states.values()])
+        X0 = cs.vcat(self._initial_states.values())
         U = cs.vcat(self._actions_exp.values())
+        D_all = cs.hcat(
+            [
+                cs.vcat([self._disturbances[_n(n, i)] for n in disturbance_names])
+                for i in range(self._n_scenarios)
+            ]
+        )
+
+        N = self._prediction_horizon
+        Fmapaccum = F.mapaccum(N, {"base": base, "allow_free": True})
+        X_next_hcat = Fmapaccum(X0, U, D_all)
+
         state_names = self.single_states.keys()
-        propagated_states = []
+        cumsizes = np.cumsum([0] + [s.shape[0] for s in self._initial_states.values()])
+        X_next_split = cs.horzsplit(X_next_hcat, N)
         for i in range(self._n_scenarios):
-            Xk_i = Xk_shared
-            D_i = cs.vcat([self._disturbances[_n(n, i)] for n in disturbance_names])
-            X_i = [Xk_i]
-            for k in range(self._prediction_horizon):
-                Xk_i = F(Xk_i, U[:, k], D_i[:, k])
-                if n_out != 1:
-                    Xk_i = Xk_i[0]
-                X_i.append(Xk_i)
-            X_i = cs.vertsplit(cs.hcat(X_i), cumsizes)
-            propagated_states.append([(_n(n, i), x) for n, x in zip(state_names, X_i)])
-        self._states = dict(chain.from_iterable(propagated_states))
+            X_i = cs.horzcat(X0, X_next_split[i])
+            X_i_split = cs.vertsplit(X_i, cumsizes)
+            for n, x in zip(state_names, X_i_split):
+                self._states[_n(n, i)] = x
+
+    def _chained_substitution_for_scenario_i(
+        self, expr: SymType, i: int
+    ) -> Union[SymType, cs.DM]:
+        """Iternal utility to perform substitutions in chain for the i-th scenario."""
+        return _chained_substitute(
+            expr,
+            (self.single_states, self.states_i(i)),
+            (self.single_disturbances, self.disturbances_i(i)),
+            (self.single_slacks, self.slacks_i(i)),
+        )
