@@ -1,22 +1,22 @@
 import random
 import unittest
-from functools import cached_property, lru_cache, partial
-from itertools import product
-from typing import Tuple, Type, Union
+from functools import cached_property, lru_cache
+from itertools import product, repeat
+from typing import Union
 
 import casadi as cs
 import numpy as np
-from casadi.tools import entry, struct_MX, struct_SX
 from parameterized import parameterized
 
+from csnlp import Nlp
 from csnlp.core.cache import invalidate_cache
-from csnlp.core.data import array2cs, cs2array
+from csnlp.core.data import array2cs, cs2array, find_index_in_vector
 from csnlp.core.debug import NlpDebug, NlpDebugEntry
 from csnlp.core.derivatives import hohessian, hojacobian
 from csnlp.core.scaling import MinMaxScaler, Scaler
-from csnlp.core.solutions import Solution, subsevalf
+from csnlp.core.solutions import EagerSolution, Solution, subsevalf
 
-GROUPS = set(NlpDebug.types.keys())
+GROUPS = set(NlpDebug._types.keys())
 
 
 class Dummy:
@@ -54,13 +54,23 @@ class Dummy2(Dummy):
         return super().clear_cache()
 
 
+class DummySolution(Solution):
+    def __init__(self, f, success, status):
+        self._f = f
+        self._stats = {"success": success, "return_status": status}
+        self._solver_plugin = "osqp"
+
+    @property
+    def f(self):
+        return self._f
+
+
 class TestFuncs(unittest.TestCase):
     def test_invalidate_cache__raises__with_invalid_type(self):
         with self.assertRaises(TypeError):
             invalidate_cache(5)
 
     def test_invalidate_cache__clears_property_cache(self):
-        # sourcery skip: extract-duplicate-method
         dummy = Dummy()
         dummy.prop1
         dummy.prop1
@@ -82,7 +92,6 @@ class TestFuncs(unittest.TestCase):
         self.assertEqual(dummy.counter2, 3)
 
     def test_invalidate_cache__invalidates_only_current_object(self):
-        # sourcery skip: extract-duplicate-method
         dummy1, dummy2 = Dummy(), Dummy()
         dummy1.prop1
         dummy2.prop1
@@ -95,7 +104,6 @@ class TestFuncs(unittest.TestCase):
         self.assertEqual(dummy2.counter1, 1)
 
     def test_invalidate_cache__accepts_new_caches_to_clear(self):
-        # sourcery skip: extract-duplicate-method
         dummy = Dummy2()
         dummy.prop1
         dummy.prop1
@@ -122,11 +130,11 @@ class TestNlpDebug(unittest.TestCase):
         shape = (2, 3)
         for group in GROUPS:
             debug.register(group, name, shape)
-            info: Tuple[range, NlpDebugEntry] = getattr(debug, f"_{group}_info")[0]
+            info: tuple[range, NlpDebugEntry] = getattr(debug, f"_{group}_info")[0]
             self.assertEqual(info[0], range(shape[0] * shape[1]))
             self.assertEqual(info[1].name, name)
             self.assertEqual(info[1].shape, shape)
-            self.assertEqual(info[1].type, NlpDebug.types[group])
+            self.assertEqual(info[1].type, NlpDebug._types[group])
 
     def test_register__raises__with_invalid_group(self):
         debug = NlpDebug()
@@ -167,23 +175,17 @@ class TestSolutions(unittest.TestCase):
         with self.assertRaises(RuntimeError):
             subsevalf(x**2, x, y, eval=True)
 
-    @parameterized.expand([(cs.SX, struct_SX), (cs.MX, struct_MX)])
+    @parameterized.expand([(cs.SX,), (cs.MX,)])
     def test_subsevalf__computes_correct_value(
-        self,
-        XX: Union[Type[cs.SX], Type[cs.MX]],
-        struct_X: Union[Type[struct_SX], Type[struct_MX]],
+        self, XX: Union[type[cs.SX], type[cs.MX]]
     ):
         shape = (3, 4)
-
         V = {
             "x": (XX.sym("x", *shape), np.random.rand(*shape) * 10),
             "y": (XX.sym("y", *shape), np.random.rand(*shape) * 5),
             "z": (XX.sym("z"), np.random.rand() + 1),
         }
-        V_vec = cs.vertcat(*(cs.vec(v) for _, v in V.values()))
-        V_struct = struct_X([entry(n, expr=s) for n, (s, _) in V.items()])
         expr, expected_val = ((V["x"][i] / V["y"][i]) ** V["z"][i] for i in range(2))
-
         actual_vals = [
             subsevalf(
                 subsevalf(
@@ -208,53 +210,222 @@ class TestSolutions(unittest.TestCase):
                 {n: v for n, (_, v) in V.items()},
                 eval=True,
             ),
-            subsevalf(expr, V_struct, V_struct(V_vec), eval=True),
         ]
         for actual_val in actual_vals:
             np.testing.assert_allclose(expected_val, actual_val)
 
-    @parameterized.expand([(cs.SX, struct_SX), (cs.MX, struct_MX)])
-    def test_solution__computes_correct_value(
-        self,
-        XX: Union[Type[cs.SX], Type[cs.MX]],
-        struct_X: Union[Type[struct_SX], Type[struct_MX]],
+    @parameterized.expand([(cs.SX,), (cs.MX,)])
+    def test_eager_solution__computes_correct_value(
+        self, XX: Union[type[cs.SX], type[cs.MX]]
     ):
-        f = 0
         shape = (3, 4)
         V = {
-            "x": (XX.sym("x", *shape), np.random.rand(*shape) * 10),
-            "y": (XX.sym("y", *shape), np.random.rand(*shape) * 5),
-            "z": (XX.sym("z"), np.random.rand() + 1),
+            "x": (XX.sym("x", shape), np.random.rand(*shape) * 10),
+            "y": (XX.sym("y", shape), np.random.rand(*shape) * 5),
+            "p": (XX.sym("p"), np.random.rand() + 1),
+        }
+        D = {
+            "g": (XX.sym("g", shape), np.random.rand(*shape) * 10),
+            "h": (XX.sym("h", shape), np.random.rand(*shape) * 5),
+            "lbx": (XX.sym("lbx", shape), np.random.rand(*shape) * 10),
+            "ubx": (XX.sym("ubx", shape), np.random.rand(*shape) * 5),
         }
-        V_vec = cs.vertcat(*(cs.vec(v) for _, v in V.values()))
-        V_struct = struct_X([entry(n, expr=s) for n, (s, _) in V.items()])
-        expr, expected_val = ((V["x"][i] / V["y"][i]) ** V["z"][i] for i in range(2))
 
-        vals = V_struct(V_vec)
-        get_value = partial(subsevalf, old=V_struct, new=vals)
-        S = Solution(
-            f=f, vars=V_struct, vals=V_struct(V_vec), stats={}, _get_value=get_value
+        def func(x, y, p, g, h, lbx, ubx):
+            return (lbx / h) * cs.exp(x * p / y + g * ubx)
+
+        expr, expected = (
+            func(
+                V["x"][i],
+                V["y"][i],
+                V["p"][i],
+                D["g"][i],
+                D["h"][i],
+                D["lbx"][i],
+                D["ubx"][i],
+            )
+            for i in range(2)
         )
-        np.testing.assert_allclose(expected_val, S.value(expr))
+        S = EagerSolution(
+            0.0,
+            V["p"][0],
+            V["p"][1],
+            cs.veccat(V["x"][0], V["y"][0]),
+            cs.veccat(V["x"][1], V["y"][1]),
+            cs.veccat(D["g"][0], D["h"][0]),
+            cs.veccat(D["g"][1], D["h"][1]),
+            cs.veccat(D["lbx"][0], D["ubx"][0]),
+            cs.veccat(D["lbx"][1], D["ubx"][1]),
+            {n: s for n, (s, _) in V.items()},
+            {n: v for n, (_, v) in V.items()},
+            {n: s for n, (s, _) in D.items()},
+            {n: v for n, (_, v) in D.items()},
+            {},
+            "a_solver_plugin",
+        )
+        np.testing.assert_allclose(expected, S.value(expr))
 
     @parameterized.expand([(False,), (True,)])
-    def test_solution__reports_success_and_barrier_properly(self, flag: bool):
+    def test_eager_solution__reports_success_and_barrier_properly(self, flag: bool):
         mu = np.abs(np.random.randn(10)).tolist()
-        S = Solution(
-            f=None,
-            vars=None,
-            vals=None,
+        S = EagerSolution(
+            *repeat(None, 13),
             stats={"success": flag, "iterations": {"mu": mu}},
-            _get_value=lambda x: x,
+            solver_plugin="a_solver_plugin",
         )
         self.assertEqual(S.success, flag)
         self.assertEqual(S.barrier_parameter, mu[-1])
 
+    def test_cmp_key__returns_correct_solution(self):
+        test_cases = [
+            {
+                "name": "All successful, varying 'f' values",
+                "sols": [(1, True, ""), (2, True, ""), (3, True, "")],
+                "expected": 0,
+            },
+            {
+                "name": "Mixed success, all feasible",
+                "sols": [(1, False, ""), (2, True, ""), (3, True, "")],
+                "expected": 1,
+            },
+            {
+                "name": "No successful, mixed feasibility",
+                "sols": [(1, False, "infeasible"), (2, False, ""), (3, False, "")],
+                "expected": 1,
+            },
+            {
+                "name": "All infeasible",
+                "sols": [
+                    (1, False, "infeasible"),
+                    (2, False, "infeasible"),
+                    (3, False, "infeasible"),
+                ],
+                "expected": 0,
+            },
+            {
+                "name": "Mixed success, infeasibility, varying 'f'",
+                "sols": [
+                    (2, True, ""),
+                    (5, True, ""),
+                    (3, False, ""),
+                    (4, False, "infeasible"),
+                ],
+                "expected": 0,
+            },
+            {
+                "name": "Real example",
+                "sols": [
+                    (387.48883666213, True, ""),
+                    (387.488836662, False, "infeasible"),
+                    (387.488836662129, True, ""),
+                    (387.4888366621303, True, ""),
+                ],
+                "expected": 2,
+            },
+        ]
+        for case in test_cases:
+            with self.subTest(name=case["name"]):
+                sols = (
+                    {"f": f, "stats": {"success": success, "return_status": status}}
+                    for f, success, status in case["sols"]
+                )
+                sols = enumerate(DummySolution(*s) for s in case["sols"])
+                min_index, _ = min(sols, key=lambda sol: DummySolution.cmp_key(sol[1]))
+                self.assertEqual(min_index, case["expected"])
+
+    @parameterized.expand(
+        product(
+            ("MX", "SX"),
+            (True, False),
+            [
+                # (
+                #     "nlp",
+                #     "sqpmethod",
+                #     {
+                #         "error_on_fail": False,
+                #         "print_time": False,
+                #         "print_status": False,
+                #         "print_header": False,
+                #         "print_iteration": False,
+                #         "qpsol_options": {
+                #             "error_on_fail": False, "printLevel": "none"
+                #         },
+                #     },
+                # ),
+                (
+                    "nlp",
+                    "ipopt",
+                    {
+                        "print_time": False,
+                        "ipopt": {"print_level": 0, "sb": "yes"},
+                    },
+                ),
+                # (
+                #     "conic",
+                #     "osqp",
+                #     {
+                #         "error_on_fail": False,
+                #         "print_time": False,
+                #         "osqp": {"verbose": False},
+                #     },
+                # ),
+                ("conic", "qpoases", {"error_on_fail": False, "printLevel": "none"}),
+                # ("conic", "proxqp", {"error_on_fail": False}),
+                (
+                    "conic",
+                    "qrqp",
+                    {
+                        "error_on_fail": False,
+                        "print_time": False,
+                        "print_header": False,
+                        "print_info": False,
+                        "print_iter": False,
+                    },
+                ),
+                ("conic", "clp", {"error_on_fail": False}),
+                (
+                    "nlp",
+                    "bonmin",
+                    {
+                        "print_time": False,
+                        "bonmin": {
+                            "fp_log_level": 0,
+                            "lp_log_level": 0,
+                            "milp_log_level": 0,
+                            "nlp_log_level": 0,
+                            "oa_cuts_log_level": 0,
+                            "oa_log_level": 0,
+                        },
+                    },
+                ),
+                # ("conic", "cbc", {"error_on_fail": False}),
+            ],
+        )
+    )
+    def test_infeasible(self, sym_type, is_feas, solver_data):
+        solver_type, solver, solver_options = solver_data
+
+        prob = Nlp(sym_type=sym_type)
+        discrete = solver in ("bonmin", "cbc", "gurobi", "knitro")
+        x, _, _ = prob.variable("x", discrete=discrete)
+        lb = -abs(np.random.randn())
+        ub = abs(np.random.randn())
+        if not is_feas:
+            lb, ub = ub, lb
+        prob.constraint("lb", x, ">=", lb)
+        prob.constraint("ub", x, "<=", ub)
+        prob.minimize(x)
+        prob.init_solver(solver_options, solver, solver_type)
+        sol = prob.solve()
+
+        self.assertEqual(sol.success, is_feas)
+        self.assertEqual(not sol.infeasible, is_feas)
+
 
 class TestData(unittest.TestCase):
     @parameterized.expand(product([cs.MX, cs.SX], [(1, 1), (3, 1), (1, 3), (3, 3)]))
     def test_cs2array_array2cs__convert_properly(
-        self, sym_type: Union[Type[cs.SX], Type[cs.MX]], shape: Tuple[int, int]
+        self, sym_type: Union[type[cs.SX], type[cs.MX]], shape: tuple[int, int]
     ):
         x = sym_type.sym("x", *shape)
         a = cs2array(x)
@@ -270,10 +441,18 @@ class TestData(unittest.TestCase):
                 o = cs.evalf(cs.substitute(y[i] - a[i], x, x_))
                 np.testing.assert_allclose(o, 0, atol=1e-9)
 
+    @parameterized.expand([(cs.SX,), (cs.MX,)])
+    def test_find_index_in_vector__works_properly(
+        self, sym_type: Union[type[cs.SX], type[cs.MX]]
+    ):
+        X = sym_type.sym("X", 5)
+        x = cs.horzcat(X[3], X[2])
+        np.testing.assert_array_equal([3, 2], find_index_in_vector(X, x))
+
 
 class TestDerivatives(unittest.TestCase):
     @parameterized.expand([((2, 2),), ((3, 1),), ((1, 3),)])
-    def test_hojacobian__computes_right_derivatives(self, shape: Tuple[int, int]):
+    def test_hojacobian__computes_right_derivatives(self, shape: tuple[int, int]):
         x = cs.SX.sym("x", *shape)
         y = (
             (x.reshape((-1, 1)) @ x.reshape((1, -1)) + (x.T if x.is_row() else x))
@@ -289,7 +468,7 @@ class TestDerivatives(unittest.TestCase):
             np.testing.assert_allclose(o, 0, atol=1e-9)
 
     @parameterized.expand([((2, 2),), ((3, 1),), ((1, 3),)])
-    def test_hohessian__computes_right_derivatives(self, shape: Tuple[int, int]):
+    def test_hohessian__computes_right_derivatives(self, shape: tuple[int, int]):
         x = cs.SX.sym("x", *shape)
         y = (
             (x.reshape((-1, 1)) @ x.reshape((1, -1)) + (x.T if x.is_row() else x))
