@@ -1,15 +1,16 @@
 import pickle
 import unittest
+from copy import deepcopy
 from itertools import product
-from typing import List, Type, Union
+from typing import Union
+from unittest.mock import Mock
 
 import casadi as cs
 import numpy as np
 from parameterized import parameterized, parameterized_class
 
-from csnlp import Nlp, ParallelMultistartNlp, StackedMultistartNlp
+from csnlp import Nlp
 from csnlp.core.solutions import subsevalf
-from csnlp.nlps.multistart_nlp import _n
 from csnlp.util.math import log
 
 OPTS = {
@@ -32,16 +33,17 @@ class TestNlp(unittest.TestCase):
         self,
         lhs: Union[cs.SX, cs.MX],
         rhs: Union[cs.SX, cs.MX],
-        vars: Union[List[cs.SX], List[cs.MX]],
+        vars: Union[list[cs.SX], list[cs.MX]],
         force_numerical: bool = False,
-    ):
+    ) -> bool:
         if isinstance(rhs, (int, float)):
             rhs = np.full(lhs.shape, rhs)
         if not force_numerical and self.sym_type == "SX":
             return cs.is_equal(lhs, rhs)
-        old = cs.vertcat(*(cs.vec(o) for o in vars))
+        old = cs.vvcat(vars)
         new = np.random.randn(*old.shape)
         np.testing.assert_allclose(subsevalf(lhs, old, new), subsevalf(rhs, old, new))
+        return True
 
     def test_init__raises__with_invalid_sym_type(self):
         with self.assertRaises(AttributeError):
@@ -83,41 +85,76 @@ class TestNlp(unittest.TestCase):
         with self.assertRaises(ValueError):
             nlp.parameter("p")
 
+    def test_variable__discrete__is_invalidated_when_new_variable_is_added(self):
+        n_variables = 30
+        np_random = np.random.default_rng()
+
+        nlp = Nlp(sym_type=self.sym_type)
+        discretes = np_random.random(n_variables) < 0.5
+        shapes = [tuple(np_random.integers(1, 10, 2)) for _ in range(n_variables)]
+        discrete_dict = {f"x{i}": d for i, d in enumerate(discretes)}
+        discrete = np.concatenate(
+            [
+                (np.ones if d else np.zeros)(np.prod(s), dtype=bool)
+                for s, d in zip(shapes, discrete_dict.values())
+            ]
+        )
+        cumnx = np.cumsum([np.prod(s) for s in shapes])
+
+        for i, (shape, discrete_) in enumerate(zip(shapes, discretes)):
+            nlp.variable(f"x{i}", shape, discrete_)
+            np.testing.assert_array_equal(nlp.discrete, discrete[: cumnx[i]])
+
     def test_variable__creates_correct_variable(self):
-        shape1 = (4, 3)
-        shape2 = (2, 2)
-        lb1, ub1 = np.random.rand(*shape1) - 1, np.random.rand(*shape1) + 1
-        lb2, ub2 = np.random.rand(*shape2) - 1, np.random.rand(*shape2) + 1
-        nx = np.prod(shape1) + np.prod(shape2)
+        n_variables = 30
+        np_random = np.random.default_rng()
+
         nlp = Nlp(sym_type=self.sym_type, debug=True)
-        x1, lam1_lb, lam1_ub = nlp.variable("x1", shape1, lb=lb1, ub=ub1)
-        x2, lam2_lb, lam2_ub = nlp.variable("x2", shape2, lb=lb2, ub=ub2)
-        for o in (x1, lam1_lb, lam1_ub):
-            self.assertEqual(o.shape, shape1)
-        for o in (x2, lam2_lb, lam2_ub):
-            self.assertEqual(o.shape, shape2)
+        discretes = np_random.random(n_variables) < 0.5
+        shapes = [tuple(np_random.integers(1, 10, 2)) for _ in range(n_variables)]
+        lbs = [np_random.random(size=s) - 1 for s in shapes]
+        ubs = [np_random.random(size=s) + 1 for s in shapes]
+        xs, lams_lb, lams_ub = [], [], []
+        for i, (shape, discrete, lb, ub) in enumerate(zip(shapes, discretes, lbs, ubs)):
+            x, lam_lb, lam_ub = nlp.variable(f"x{i}", shape, discrete, lb, ub)
+            xs.append(x)
+            lams_lb.append(lam_lb)
+            lams_ub.append(lam_ub)
+
+        for shape, x, lam_lb, lam_ub in zip(shapes, xs, lams_lb, lams_ub):
+            self.assertEqual(x.shape, shape)
+            self.assertEqual(lam_lb.shape, (np.prod(shape), 1))
+            self.assertEqual(lam_ub.shape, (np.prod(shape), 1))
+
+        nx = sum(np.prod(s) for s in shapes)
         self.assertEqual(nlp.nx, nx)
+        self.cmp(nlp.x, cs.vvcat(xs), vars=xs)
 
-        x = cs.vertcat(cs.vec(x1), cs.vec(x2))
-        self.cmp(nlp.x, x, vars=[x1, x2])
+        discrete = np.concatenate(
+            [
+                (np.ones if d else np.zeros)(np.prod(s), dtype=bool)
+                for s, d in zip(shapes, discretes)
+            ]
+        )
+        np.testing.assert_array_equal(nlp.discrete, discrete)
 
-        lb = cs.vertcat(cs.vec(lb1), cs.vec(lb2))
-        ub = cs.vertcat(cs.vec(ub1), cs.vec(ub2))
+        lb = cs.vvcat(lbs)
         np.testing.assert_allclose(nlp.lbx, lb.full().flat)
+        ub = cs.vvcat(ubs)
         np.testing.assert_allclose(nlp.ubx, ub.full().flat)
 
-        i = 0
-        for name, shape in [("x1", shape1), ("x2", shape2)]:
+        k = 0
+        for i, shape in enumerate(shapes):
+            name = f"x{i}"
             for _ in range(np.prod(shape)):
-                self.assertEqual(name, nlp.debug.x_describe(i).name)
-                self.assertEqual(shape, nlp.debug.x_describe(i).shape)
-                i += 1
+                self.assertEqual(name, nlp.debug.x_describe(k).name)
+                self.assertEqual(shape, nlp.debug.x_describe(k).shape)
+                k += 1
+            self.assertTrue(cs.is_equal(nlp.variables[name], xs[i]))
+
         with self.assertRaises(IndexError):
             nlp.debug.x_describe(nx + 1)
 
-        self.assertTrue(cs.is_equal(nlp.variables["x1"], x1))
-        self.assertTrue(cs.is_equal(nlp.variables["x2"], x2))
-
     def test_variable__raises__with_variables_with_same_name(self):
         nlp = Nlp(sym_type=self.sym_type)
         nlp.variable("x")
@@ -164,15 +201,18 @@ class TestNlp(unittest.TestCase):
     @parameterized.expand([("==", "=="), ("==", ">="), (">=", ">=")])
     def test_constraint__creates_constraint_correctly(self, op1: str, op2: str):
         shape1, shape2 = (4, 3), (2, 2)
-        nc = np.prod(shape1) + np.prod(shape2)
+        nc1, nc2 = np.prod(shape1), np.prod(shape2)
+        nc = nc1 + nc2
         nlp = Nlp(sym_type=self.sym_type, debug=True)
         x = nlp.variable("x", shape1)[0]
         y = nlp.variable("y", shape2)[0]
         c1, lam_c1 = nlp.constraint("c1", x, op1, 5)
         c2, lam_c2 = nlp.constraint("c2", 5, op2, y)
 
-        self.assertTrue(c1.shape == lam_c1.shape == shape1)
-        self.assertTrue(c2.shape == lam_c2.shape == shape2)
+        self.assertEqual(c1.shape, shape1)
+        self.assertEqual(lam_c1.shape, (nc1, 1))
+        self.assertEqual(c2.shape, shape2)
+        self.assertEqual(lam_c2.shape, (nc2, 1))
         self.assertEqual(nlp.ng + nlp.nh, nc)
 
         i, prev_op = 0, op1
@@ -188,8 +228,8 @@ class TestNlp(unittest.TestCase):
             with self.assertRaises(IndexError):
                 describe(nc + 1)
 
-        c = cs.vertcat(cs.vec(c1), cs.vec(c2))
-        lam = cs.vertcat(cs.vec(lam_c1), cs.vec(lam_c2))
+        c = cs.veccat(c1, c2)
+        lam = cs.vertcat(lam_c1, lam_c2)
         expected_c = cs.vertcat(nlp.g, nlp.h)
         expected_lam = cs.vertcat(nlp.lam_g, nlp.lam_h)
         self.cmp(expected_c, c, vars=[x, y])
@@ -203,7 +243,6 @@ class TestNlp(unittest.TestCase):
         self.assertIn("slack_c0", nlp._vars)
 
     def test_constraint__solves_correctly__with_soft_variable(self):
-        # sourcery skip: extract-duplicate-method, inline-immediately-returned-variable
         # From https://www.gams.com/latest/docs/UG_EMP_SoftConstraints.html
         # solve with manual soft variable
         nlp = Nlp(sym_type=self.sym_type)
@@ -211,7 +250,7 @@ class TestNlp(unittest.TestCase):
         v = nlp.variable("v", lb=0)[0]
         nlp.constraint("c0", 3 * x[0] + x[1], "<=", 5)
         nlp.constraint("c1", v, ">=", 20 - x[1] ** 2)
-        nlp.minimize(-x[0] ** 2 + 5 * (cs.log(x[0]) - 1) ** 2 + 2 * v)
+        nlp.minimize(-(x[0] ** 2) + 5 * (cs.log(x[0]) - 1) ** 2 + 2 * v)
         nlp.init_solver(OPTS)
         sol1 = nlp.solve(vals0={"x": [1, 1], "v": 0})
 
@@ -220,7 +259,7 @@ class TestNlp(unittest.TestCase):
         x = nlp.variable("x", (2, 1), lb=0)[0]
         nlp.constraint("c0", 3 * x[0] + x[1], "<=", 5)
         _, _, v = nlp.constraint("c1", x[1] ** 2, ">=", 20, soft=True)
-        nlp.minimize(-x[0] ** 2 + 5 * (cs.log(x[0]) - 1) ** 2 + 2 * v)
+        nlp.minimize(-(x[0] ** 2) + 5 * (cs.log(x[0]) - 1) ** 2 + 2 * v)
         nlp.init_solver(OPTS)
         sol2 = nlp.solve(vals0={"x": [1, 1], "slack_c1": 0})
 
@@ -250,8 +289,8 @@ class TestNlp(unittest.TestCase):
         _, lam_c2 = nlp.constraint("c2", 5, op2, y)
 
         dv = nlp.dual_variables
-        c1_name = f'lam_{"g" if op1 == "==" else "h"}_c1'
-        c2_name = f'lam_{"g" if op2 == "==" else "h"}_c2'
+        c1_name = f"lam_{'g' if op1 == '==' else 'h'}_c1"
+        c2_name = f"lam_{'g' if op2 == '==' else 'h'}_c2"
         self.assertTrue(cs.is_equal(dv["lam_lb_x"], lam_lb_x))
         self.assertTrue(cs.is_equal(dv["lam_ub_x"], lam_ub_x))
         self.assertTrue(cs.is_equal(dv["lam_lb_y"], lam_lb_y))
@@ -272,155 +311,95 @@ class TestNlp(unittest.TestCase):
         x, lam_lbx, lam_ubx = nlp.variable("x", (2, 3), lb=[[0], [-np.inf]], ub=1)
         _, lam_g = nlp.constraint("c1", x[:, 0], "==", 2)
         _, lam_h = nlp.constraint("c2", x[0, :] + x[1, :] ** 2, "<=", 2)
-        v = cs.vec
-        vc = cs.vertcat
 
-        lam = vc(v(lam_g), v(lam_h), v(lam_lbx[0, :]), v(lam_ubx))
-        lam_all = vc(v(lam_g), v(lam_h), v(lam_lbx), v(lam_ubx))
-        y = vc(v(x), v(lam_g), v(lam_h), v(lam_lbx[0, :]), v(lam_ubx))
-        y_all = vc(v(x), v(lam_g), v(lam_h), v(lam_lbx), v(lam_ubx))
+        lam = cs.vertcat(lam_g, lam_h, lam_lbx, lam_ubx)
+        y = cs.veccat(x, lam_g, lam_h, lam_lbx, lam_ubx)
 
         vars = [x, lam_g, lam_h, lam_lbx, lam_ubx]
         self.cmp(nlp.lam, lam, vars=vars)
-        self.cmp(nlp.lam_all, lam_all, vars=vars)
-        self.cmp(nlp.primal_dual_vars(all=False), y, vars=vars)
-        self.cmp(nlp.primal_dual_vars(all=True), y_all, vars=vars)
+        self.cmp(nlp.primal_dual, y, vars=vars)
 
     @parameterized.expand([(True,), (False,)])
-    def test_h_lbx_ubx__returns_correct_indices(self, flag: bool):
-        # sourcery skip: low-code-quality
-        nlp = Nlp(sym_type=self.sym_type, remove_redundant_x_bounds=flag)
+    def test_h_lbx_ubx__returns_correct_indices(self, remove_red_bnds: bool):
+        nlp = Nlp(sym_type=self.sym_type, remove_redundant_x_bounds=remove_red_bnds)
 
-        x1, lam_lbx1, lam_ubx1 = nlp.variable("x1", (2, 1))
-        (h_lbx, lam_lbx), (h_ubx, lam_ubx) = nlp.h_lbx, nlp.h_ubx
-        vars = [x1, lam_lbx1, lam_ubx1]
-        if flag:
-            self.assertTrue(all(o.is_empty() for o in [h_lbx, lam_lbx, h_ubx, lam_ubx]))
-        else:
-            if self.sym_type == "SX":
-                self.cmp((h_lbx - (-np.inf - x1)), 0, vars=vars, force_numerical=True)
-                self.cmp((h_ubx - (x1 - np.inf)), 0, vars=vars, force_numerical=True)
-            self.cmp(cs.evalf(lam_lbx - lam_lbx1), 0, vars=vars)
-            self.cmp(cs.evalf(lam_ubx - lam_ubx1), 0, vars=vars)
+        x1, lam_lbx_c1, lam_ubx_c1 = nlp.variable("x1", (2, 1))
+        lam_lbx1, lam_ubx1 = nlp.lam_lbx, nlp.lam_ubx
+        h_lbx1, h_ubx1 = nlp.h_lbx, nlp.h_ubx
+        vars1 = [x1, lam_lbx_c1, lam_ubx_c1]
+        x2, lam_lbx_c2, lam_ubx_c2 = nlp.variable("x2", (2, 1), lb=0)
+        lam_lbx2, lam_ubx2 = nlp.lam_lbx, nlp.lam_ubx
+        h_lbx2, h_ubx2 = nlp.h_lbx, nlp.h_ubx
+        vars2 = vars1.copy() + [x2, lam_lbx_c2, lam_ubx_c2]
+        x3, lam_lbx_c3, lam_ubx_c3 = nlp.variable("x3", (2, 1), ub=1)
+        lam_lbx3, lam_ubx3 = nlp.lam_lbx, nlp.lam_ubx
+        h_lbx3, h_ubx3 = nlp.h_lbx, nlp.h_ubx
+        vars3 = vars2.copy() + [x3, lam_lbx_c3, lam_ubx_c3]
+        x4, lam_lbx_c4, lam_ubx_c4 = nlp.variable("x4", (2, 1), lb=0, ub=1)
+        lam_lbx4, lam_ubx4 = nlp.lam_lbx, nlp.lam_ubx
+        h_lbx4, h_ubx4 = nlp.h_lbx, nlp.h_ubx
+        vars4 = vars3.copy() + [x4, lam_lbx_c4, lam_ubx_c4]
 
-        x2, lam_lbx2, lam_ubx2 = nlp.variable("x2", (2, 1), lb=0)
-        (h_lbx, lam_lbx), (h_ubx, lam_ubx) = nlp.h_lbx, nlp.h_ubx
-        vars.extend([x2, lam_lbx2, lam_ubx2])
-        if flag:
-            self.assertTrue(all(o.is_empty() for o in [h_ubx, lam_ubx]))
-            self.cmp(cs.evalf(h_lbx - (-x2)), 0, vars=vars)
-            self.cmp(cs.evalf(lam_lbx - lam_lbx2), 0, vars=vars)
+        if remove_red_bnds:
+            self.assertTrue(
+                all(o.is_empty() for o in [h_lbx1, lam_lbx1, h_ubx1, lam_ubx1])
+            )
+            self.assertTrue(all(o.is_empty() for o in [h_ubx2, lam_ubx2]))
+            self.cmp(cs.evalf(h_lbx2 - (-x2)), 0, vars=vars2)
+            self.cmp(cs.evalf(lam_lbx2 - lam_lbx_c2), 0, vars=vars2)
+            self.cmp(cs.evalf(h_lbx3 - (-x2)), 0, vars=vars3)
+            self.cmp(cs.evalf(lam_lbx3 - lam_lbx2), 0, vars=vars3)
+            self.cmp((h_ubx3 - (x3 - 1)), 0, vars=vars3, force_numerical=True)
+            self.cmp(cs.evalf(lam_ubx3 - lam_ubx_c3), 0, vars=vars3)
+            self.cmp(h_lbx4 - cs.vertcat(-x2, -x4), 0, vars=vars4, force_numerical=True)
+            e = lam_lbx4 - cs.vertcat(lam_lbx_c2, lam_lbx_c4)
+            self.cmp(e, 0, vars=vars4, force_numerical=True)
+            e = h_ubx4 - cs.vertcat(x3 - 1, x4 - 1)
+            self.cmp(e, 0, vars=vars4, force_numerical=True)
+            e = lam_ubx4 - cs.vertcat(lam_ubx_c3, lam_ubx_c4)
+            self.cmp(e, 0, vars=vars4, force_numerical=True)
         else:
-            if self.sym_type == "SX":
-                self.cmp(
-                    cs.evalf(h_lbx - cs.vertcat(-np.inf - x1, 0 - x2)),
-                    0,
-                    vars=vars,
-                )
-                self.cmp(
-                    cs.evalf(h_ubx - cs.vertcat(x1 - np.inf, x2 - np.inf)),
-                    0,
-                    vars=vars,
-                )
-            self.cmp(
-                (lam_lbx - cs.vertcat(lam_lbx1, lam_lbx2)),
-                0,
-                vars=vars,
-                force_numerical=True,
-            )
-            self.cmp(
-                (lam_ubx - cs.vertcat(lam_ubx1, lam_ubx2)),
-                0,
-                vars=vars,
-            )
+            self.cmp(cs.evalf(lam_lbx1 - lam_lbx_c1), 0, vars=vars1)
+            self.cmp(cs.evalf(lam_ubx1 - lam_ubx_c1), 0, vars=vars1)
+            e = lam_lbx2 - cs.vertcat(lam_lbx_c1, lam_lbx_c2)
+            self.cmp(e, 0, vars=vars2, force_numerical=True)
+            e = lam_ubx2 - cs.vertcat(lam_ubx_c1, lam_ubx_c2)
+            self.cmp(e, 0, vars=vars2)
+            e = lam_lbx3 - cs.vertcat(lam_lbx_c1, lam_lbx_c2, lam_lbx_c3)
+            self.cmp(e, 0, vars=vars3, force_numerical=True)
+            e = lam_ubx3 - cs.vertcat(lam_ubx_c1, lam_ubx_c2, lam_ubx_c3)
+            self.cmp(e, 0, vars=vars3, force_numerical=True)
+            e = lam_lbx4 - cs.vertcat(lam_lbx_c1, lam_lbx_c2, lam_lbx_c3, lam_lbx_c4)
+            self.cmp(e, 0, vars=vars4, force_numerical=True)
+            e = lam_ubx4 - cs.vertcat(lam_ubx_c1, lam_ubx_c2, lam_ubx_c3, lam_ubx_c4)
+            self.cmp(e, 0, vars=vars4, force_numerical=True)
 
-        x3, lam_lbx3, lam_ubx3 = nlp.variable("x3", (2, 1), ub=1)
-        (h_lbx, lam_lbx), (h_ubx, lam_ubx) = nlp.h_lbx, nlp.h_ubx
-        vars.extend([x3, lam_lbx3, lam_ubx3])
-        if flag:
-            self.cmp(cs.evalf(h_lbx - (-x2)), 0, vars=vars)
-            self.cmp(cs.evalf(lam_lbx - lam_lbx2), 0, vars=vars)
-            self.cmp((h_ubx - (x3 - 1)), 0, vars=vars, force_numerical=True)
-            self.cmp(cs.evalf(lam_ubx - lam_ubx3), 0, vars=vars)
-        else:
             if self.sym_type == "SX":
-                self.cmp(
-                    cs.evalf(h_lbx - cs.vertcat(-np.inf - x1, 0 - x2, -np.inf - x3)),
-                    0,
-                    vars=vars,
-                )
-                self.cmp(
-                    cs.evalf(h_ubx - cs.vertcat(x1 - np.inf, x2 - np.inf, x3 - 1)),
-                    0,
-                    vars=vars,
-                )
-            self.cmp(
-                (lam_lbx - cs.vertcat(lam_lbx1, lam_lbx2, lam_lbx3)),
-                0,
-                vars=vars,
-                force_numerical=True,
-            )
-            self.cmp(
-                (lam_ubx - cs.vertcat(lam_ubx1, lam_ubx2, lam_ubx3)),
-                0,
-                vars=vars,
-                force_numerical=True,
-            )
-
-        x4, lam_lbx4, lam_ubx4 = nlp.variable("x4", (2, 1), lb=0, ub=1)
-        (h_lbx, lam_lbx), (h_ubx, lam_ubx) = nlp.h_lbx, nlp.h_ubx
-        vars.extend([x4, lam_lbx4, lam_ubx4])
-        if flag:
-            self.cmp((h_lbx - cs.vertcat(-x2, -x4)), 0, vars=vars, force_numerical=True)
-            self.cmp(
-                (lam_lbx - cs.vertcat(lam_lbx2, lam_lbx4)),
-                0,
-                vars=vars,
-                force_numerical=True,
-            )
-            self.cmp(
-                (h_ubx - cs.vertcat(x3 - 1, x4 - 1)), 0, vars=vars, force_numerical=True
-            )
-            self.cmp(
-                (lam_ubx - cs.vertcat(lam_ubx3, lam_ubx4)),
-                0,
-                vars=vars,
-                force_numerical=True,
-            )
-        else:
-            if self.sym_type == "SX":
-                self.cmp(
-                    cs.evalf(
-                        h_lbx - cs.vertcat(-np.inf - x1, 0 - x2, -np.inf - x3, 0 - x4)
-                    ),
-                    0,
-                    vars=vars,
-                )
-                self.cmp(
-                    cs.evalf(
-                        h_ubx - cs.vertcat(x1 - np.inf, x2 - np.inf, x3 - 1, x4 - 1)
-                    ),
-                    0,
-                    vars=vars,
-                )
-            self.cmp(
-                (lam_lbx - cs.vertcat(lam_lbx1, lam_lbx2, lam_lbx3, lam_lbx4)),
-                0,
-                vars=vars,
-                force_numerical=True,
-            )
-            self.cmp(
-                (lam_ubx - cs.vertcat(lam_ubx1, lam_ubx2, lam_ubx3, lam_ubx4)),
-                0,
-                vars=vars,
-                force_numerical=True,
-            )
+                self.cmp(h_lbx1 - (-np.inf - x1), 0, vars=vars1, force_numerical=True)
+                self.cmp(h_ubx1 - (x1 - np.inf), 0, vars=vars1, force_numerical=True)
+                e = h_lbx2 - cs.vertcat(-np.inf - x1, 0 - x2)
+                self.cmp(cs.evalf(e), 0, vars=vars2)
+                e = h_ubx2 - cs.vertcat(x1 - np.inf, x2 - np.inf)
+                self.cmp(cs.evalf(e), 0, vars=vars2)
+                e = h_lbx3 - cs.vertcat(-np.inf - x1, 0 - x2, -np.inf - x3)
+                self.cmp(cs.evalf(e), 0, vars=vars3)
+                e = h_ubx3 - cs.vertcat(x1 - np.inf, x2 - np.inf, x3 - 1)
+                self.cmp(cs.evalf(e), 0, vars=vars3)
+                e = h_lbx4 - cs.vertcat(-np.inf - x1, 0 - x2, -np.inf - x3, 0 - x4)
+                self.cmp(cs.evalf(e), 0, vars=vars4)
+                e = h_ubx4 - cs.vertcat(x1 - np.inf, x2 - np.inf, x3 - 1, x4 - 1)
+                self.cmp(cs.evalf(e), 0, vars=vars4)
 
     def test_init_solver__raises__when_objective_not_set(self):
         nlp = Nlp(sym_type=self.sym_type)
         with self.assertRaises(RuntimeError):
             nlp.init_solver(OPTS)
 
+    def test_init_solver__raises__when_type_is_wrong(self):
+        nlp = Nlp(sym_type=self.sym_type)
+        with self.assertRaises(ValueError):
+            nlp.init_solver(OPTS, type="a_random_type")
+
     def test_init_solver__saves_options_correctly(self):
         nlp = Nlp(sym_type=self.sym_type)
         x = nlp.variable("x")[0]
@@ -428,6 +407,27 @@ class TestNlp(unittest.TestCase):
         nlp.init_solver(OPTS)
         self.assertDictEqual(OPTS, nlp.solver_opts)
 
+    @parameterized.expand([("sqpmethod",), ("osqp",)])
+    def test_init_solver__chooses_conic_or_nlp_correctly(self, solver: str):
+        nlp = Nlp(sym_type=self.sym_type)
+        x = nlp.variable("x")[0]
+        nlp.minimize(x**2)
+
+        old_qpsol, old_nlpsol = cs.qpsol, cs.nlpsol
+        cs.qpsol = mock_qpsol = Mock()
+        cs.nlpsol = mock_nlpsol = Mock()
+        try:
+            nlp.init_solver(solver=solver)
+        finally:
+            cs.qpsol, cs.nlpsol = old_qpsol, old_nlpsol
+
+        if solver == "sqpmethod":
+            mock_qpsol.assert_not_called()
+            mock_nlpsol.assert_called_once()
+        else:
+            mock_qpsol.assert_called_once()
+            mock_nlpsol.assert_not_called()
+
     def test_solve__raises__with_uninit_solver(self):
         nlp = Nlp(sym_type=self.sym_type)
         with self.assertRaisesRegex(RuntimeError, "Solver uninitialized."):
@@ -454,13 +454,13 @@ class TestNlp(unittest.TestCase):
         sol = nlp.solve({"p": 3})
         self.assertTrue(sol.success)
         np.testing.assert_allclose(sol.f, 3)
-        for k in sol.vals.keys():
-            np.testing.assert_allclose(sol.vals[k], 0)
+        for v in sol.vals.values():
+            np.testing.assert_allclose(v, 0)
         o = sol.value(p + (x.T @ x + y.T @ y))
         np.testing.assert_allclose(sol.f, o)
 
     def test_solve__computes_corretly__example_1a(self):
-        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1a
+        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1
         nlp = Nlp(sym_type=self.sym_type)
         x = nlp.variable("x")[0]
         y = nlp.variable("y")[0]
@@ -474,7 +474,7 @@ class TestNlp(unittest.TestCase):
         np.testing.assert_allclose(sol.value(nlp.lam_g), 1 / np.sqrt(2), atol=1e-9)
 
     def test_solve__computes_corretly__example_1b(self):
-        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1b
+        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1
         nlp = Nlp(sym_type=self.sym_type)
         x = nlp.variable("x")[0]
         y = nlp.variable("y")[0]
@@ -504,7 +504,7 @@ class TestNlp(unittest.TestCase):
         )
 
     def test_solve__computes_corretly__example_3(self):
-        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_3:_Entropy
+        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_3
         n = 50
         nlp = Nlp(sym_type=self.sym_type)
         p = nlp.variable("p", (n, 1), lb=1 / n * 1e-6)[0]
@@ -518,7 +518,7 @@ class TestNlp(unittest.TestCase):
         )
 
     def test_solve__computes_corretly__example_4(self):
-        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_4:_Numerical_optimization
+        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_4
         nlp = Nlp(sym_type=self.sym_type)
         x = nlp.variable("x")[0]
         nlp.constraint("c1", x**2, "==", 1)
@@ -558,7 +558,7 @@ class TestNlp(unittest.TestCase):
         nlp.constraint("c3", z.T @ [15, 22, 4] + 3 * cs.sqrt(cs.sumsqr(z)), "<=", 38)
         nlp.init_solver(OPTS)
         sol = nlp.solve()
-        for (name, v) in [("x", x), ("y", y), ("z", z)]:
+        for name, v in [("x", x), ("y", y), ("z", z)]:
             val1 = sol.vals[name]
             val2 = sol.value(v)
             val3 = sol.value(nlp.variables[name])
@@ -586,9 +586,9 @@ class TestNlp(unittest.TestCase):
         p = nlp.parameter("p")
         nlp.minimize((1 - x) ** 2 + a * (y - x**2) ** 2)
         nlp.init_solver(OPTS)
-        with self.assertRaises(ValueError):
+        with self.assertRaises(RuntimeError):
             nlp.to_function("M", [p], [xy], ["xy"], ["xy"])
-        with self.assertRaises(ValueError):
+        with self.assertRaises(RuntimeError):
             nlp.to_function("M", [p, xy], [xy, c], ["xy"], ["xy"])
 
     @parameterized.expand([(False,), (True,)])
@@ -598,7 +598,7 @@ class TestNlp(unittest.TestCase):
         a = 0.2
         nlp = Nlp(sym_type=self.sym_type)
         if copy:
-            nlp = nlp.copy()
+            nlp = deepcopy(nlp)
         x = nlp.variable("x", lb=0)[0]
         y = nlp.variable("y")[0]
         xy = cs.vertcat(x, y)
@@ -609,7 +609,7 @@ class TestNlp(unittest.TestCase):
         nlp.constraint("c2", g, "<=", p**2)
         nlp.init_solver(OPTS)
 
-        M = nlp.to_function("M", [p, xy], [xy], ["p", "xy"], ["xy"])
+        M = nlp.to_function("M", [p, xy], [xy], ["p", "xy"], ["xy_opt"])
 
         sol = nlp.solve(pars={"p": 1.25})
         xy1 = sol.value(xy).full().flatten()
@@ -630,112 +630,130 @@ class TestNlp(unittest.TestCase):
         nlp.constraint("c2", g, "<=", p**2)
         nlp.init_solver(OPTS)
 
-        nlp2 = pickle.loads(pickle.dumps(nlp))
+        with cs.global_pickle_context():
+            pickled = pickle.dumps(nlp)
+        with cs.global_unpickle_context():
+            other = pickle.loads(pickled)
 
-        self.assertEqual(nlp.name, nlp2.name)
+        self.assertEqual(nlp.name, other.name)
 
+    @parameterized.expand(product(("both", "lb", "ub"), (True, False)))
+    def test_remove_variable_bounds__remove_bounds_correctly(
+        self, direction: str, all_idx: bool
+    ):
+        shape = tuple(np.random.randint(3, 10, size=2))
+        lb = np.random.rand(*shape) - 3
+        ub = np.random.rand(*shape) + 3
+        if all_idx:
+            idx_to_remove = list(product(range(shape[0]), range(shape[1])))
+        else:
+            n_to_remove = np.random.randint(1, np.prod(shape) // 2)
+            idx_to_remove = np.random.randint((0, 0), shape, size=(n_to_remove, 2))
 
-@parameterized_class("sym_type", [("SX",), ("MX",)])
-class TestMultistartNlp(unittest.TestCase):
-    def test_init__raises__with_invalid_number_of_starts(self):
-        with self.assertRaisesRegex(
-            ValueError, "Number of scenarios must be positive and > 0."
-        ):
-            StackedMultistartNlp(starts=0, sym_type=self.sym_type)
+        nlp = Nlp(sym_type=self.sym_type, remove_redundant_x_bounds=True)
+        u_size = np.prod(nlp.variable("u", (5, 2), ub=+1)[0].shape)  # to create noise
+        nlp.variable("x", shape, lb=lb, ub=ub)
+        nlp.variable("z", (7, 9), lb=+2, ub=+3)  # to create noise
+        nlp.remove_variable_bounds("x", direction, None if all_idx else idx_to_remove)
 
-    def test_variable_parameter_and_constraint__builds_correct_copies(self):
-        N = 3
-        nlp = StackedMultistartNlp(starts=N, sym_type=self.sym_type)
-        x = nlp.variable("x")[0]
-        y = nlp.variable("y")[0]
-        z = nlp.variable("z")[0]
-        p = nlp.parameter("p")
-        nlp.constraint("c0", x**2, "<=", p)
-        nlp.constraint("c1", y, "==", p)
-        nlp.constraint("c2", z, ">=", -p)
-        self.assertEqual(nlp.nx, nlp._stacked_nlp.nx // N)
-        self.assertEqual(nlp.np, nlp._stacked_nlp.np // N)
-        self.assertEqual(nlp.ng, nlp._stacked_nlp.ng // N)
-        self.assertEqual(nlp.nh, nlp._stacked_nlp.nh // N)
+        if direction in {"both", "lb"}:
+            lb_ = lb.copy()
+            lb_mask_ = np.full(lb.shape, False)
+            for i in idx_to_remove:
+                lb_[tuple(i)] = -np.inf
+                lb_mask_[tuple(i)] = True
+            exp_lb = nlp.lbx.data[u_size : u_size + np.prod(shape)]
+            exp_lb_mask = nlp.lbx.mask[u_size : u_size + np.prod(shape)]
+            np.testing.assert_array_equal(lb_.reshape(-1, order="F"), exp_lb)
+            np.testing.assert_array_equal(lb_mask_.reshape(-1, order="F"), exp_lb_mask)
+            self.assertTrue(
+                nlp.dual_variables["lam_lb_x"].size1() == (~exp_lb_mask).sum()
+            )
+            self.assertTrue(nlp.h_lbx.shape == nlp.lam_lbx.shape)
+        if direction in {"both", "ub"}:
+            ub_ = ub.copy()
+            ub_mask_ = np.full(ub.shape, False)
+            for i in idx_to_remove:
+                ub_[tuple(i)] = +np.inf
+                ub_mask_[tuple(i)] = True
+            exp_ub = nlp.ubx.data[u_size : u_size + np.prod(shape)]
+            exp_ub_mask = nlp.ubx.mask[u_size : u_size + np.prod(shape)]
+            np.testing.assert_array_equal(ub_.reshape(-1, order="F"), exp_ub)
+            np.testing.assert_array_equal(ub_mask_.reshape(-1, order="F"), exp_ub_mask)
+            self.assertTrue(
+                nlp.dual_variables["lam_ub_x"].size1() == (~exp_ub_mask).sum()
+            )
+            self.assertTrue(nlp.h_ubx.shape == nlp.lam_ubx.shape)
 
-    def test_minimize__sums_objectives_in_unique_function(self):
-        N = 3
-        nlp = StackedMultistartNlp(starts=N, sym_type=self.sym_type)
-        x = nlp.variable("x")[0]
-        nlp.minimize(cs.exp((x - 1) ** 2))
-        x_ = cs.DM(np.random.randn(*x.shape))
-        x_dict = {_n("x", i): x_ for i in range(N)}
-        f1 = subsevalf(nlp.f, x, x_)
-        f2 = subsevalf(nlp._stacked_nlp.f, nlp._stacked_nlp._vars, x_dict)
-        self.assertAlmostEqual(f1, f2 / N)
-
-    def test_solve__raises__with_both_flags_on(self):
-        N = 3
-        nlp = StackedMultistartNlp(starts=N, sym_type=self.sym_type)
-        with self.assertRaisesRegex(
-            AssertionError,
-            "`return_all_sols` and `return_stacked_sol` can't be both true.",
-        ):
-            nlp(None, None, return_all_sols=True, return_stacked_sol=True)
-
-    @parameterized.expand(
-        product([False, True], [StackedMultistartNlp, ParallelMultistartNlp])
-    )
-    def test_solve__computes_right_solution(self, copy: bool, multinlp_cls: Type):
-        N = 3
-        nlp = multinlp_cls(starts=N, sym_type=self.sym_type)
-        x = nlp.variable("x", lb=-0.5, ub=1.4)[0]
-        nlp.parameter("p")
-        nlp.minimize(
-            -0.3 * x**2
-            - cs.exp(-10 * x**2)
-            + cs.exp(-100 * (x - 1) ** 2)
-            + cs.exp(-100 * (x - 1.5) ** 2)
-        )
+    @parameterized.expand(product(("both", "g", "h"), (False, True)))
+    def test_remove_constraints__remove_bounds_correctly(
+        self, remove: str, all_idx: bool
+    ):
+        nlp = Nlp(sym_type=self.sym_type)
+        x = nlp.variable("x", tuple(np.random.randint(2, 5, size=2)))[0]
+        y = nlp.variable("y", tuple(np.random.randint(2, 5, size=2)))[0]
+        z = nlp.variable("z", tuple(np.random.randint(2, 5, size=2)))[0]  # noise
+        w = nlp.variable("w", tuple(np.random.randint(2, 5, size=2)))[0]  # noise
+        q = nlp.variable("q", tuple(np.random.randint(2, 5, size=2)))[0]  # noise
+        p = nlp.variable("p", tuple(np.random.randint(2, 5, size=2)))[0]  # noise
+        nlp.minimize(cs.sumsqr(cs.veccat(x, y, z, w, q, p)))
+        nlp.constraint("h0", z, ">=", 0.0)  # noise
+        nlp.constraint("h1", x, ">=", 1.0)
+        nlp.constraint("h2", q, ">=", 0.0)  # noise
+        nlp.constraint("g0", w, "==", 0.0)  # noise
+        nlp.constraint("g1", y, "==", 2.0)
+        nlp.constraint("g2", p, "==", 0.0)  # noise
         nlp.init_solver(OPTS)
-        if copy:
-            nlp = nlp.copy()
 
-        # solve manually
-        x0s = [0.9, 0.5, 1.1]
-        xfs, fs = [], []
-        for x0 in x0s:
-            sol = nlp.solve(pars={"p": 0}, vals0={"x": x0})
-            xfs.append(sol.vals["x"])
-            fs.append(sol.f)
+        if all_idx:
+            idx_to_remove_h1 = idx_to_remove_g1 = None
+        else:
+            n_to_remove = np.random.randint(1, np.prod(x.shape) // 2)
+            idx_to_remove_h1 = np.unique(
+                np.random.randint((0, 0), x.shape, size=(n_to_remove, 2)), axis=0
+            )
+            n_to_remove = np.random.randint(1, np.prod(y.shape) // 2)
+            idx_to_remove_g1 = np.unique(
+                np.random.randint((0, 0), y.shape, size=(n_to_remove, 2)), axis=0
+            )
+        remove_h = remove in {"both", "h"}
+        remove_g = remove in {"both", "g"}
+        if remove_h:
+            nlp.remove_constraints("h1", idx_to_remove_h1)
+        if remove_g:
+            nlp.remove_constraints("g1", idx_to_remove_g1)
 
-        # solve with multistart
-        args = ({"p": 0}, [{"x": x0} for x0 in x0s])
-        best_sol = nlp.solve_multi(*args)
-        all_sols = nlp.solve_multi(*args, return_all_sols=True)
+        sol = nlp.solve()
 
-        for xf, f, sol in zip(xfs, fs, all_sols):
-            np.testing.assert_allclose(xf, sol.vals["x"])
-            np.testing.assert_allclose(xf, sol.value(x))
-            np.testing.assert_allclose(f, sol.f)
-            np.testing.assert_allclose(f, sol.value(nlp.f))
-        np.testing.assert_allclose(best_sol.f, min(fs))
-        np.testing.assert_allclose(best_sol.value(nlp.f), min(fs))
+        expected_nh = cs.veccat(z, x, q).size1()
+        expected_ng = cs.veccat(w, y, p).size1()
+        expected_x = np.full(x.shape, 1.0)
+        expected_y = np.full(y.shape, 2.0)
+        if remove_h:
+            if all_idx:
+                expected_nh -= np.prod(x.shape)
+                expected_x.fill(0.0)
+            else:
+                expected_nh -= idx_to_remove_h1.shape[0]
+                for idx in idx_to_remove_h1:
+                    expected_x[idx[0], idx[1]] = 0.0
+        if remove_g:
+            if all_idx:
+                expected_ng -= np.prod(y.shape)
+                expected_y.fill(0.0)
+            else:
+                expected_ng -= idx_to_remove_g1.shape[0]
+                for idx in idx_to_remove_g1:
+                    expected_y[idx[0], idx[1]] = 0.0
 
-    @parameterized.expand([(StackedMultistartNlp,), (ParallelMultistartNlp,)])
-    def test_is_pickleable(self, multinlp_cls: Type):
-        N = 3
-        nlp = multinlp_cls(starts=N, sym_type=self.sym_type)
-        x = nlp.variable("x", lb=-0.5, ub=1.4)[0]
-        nlp.parameter("p")
-        nlp.minimize(
-            -0.3 * x**2
-            - cs.exp(-10 * x**2)
-            + cs.exp(-100 * (x - 1) ** 2)
-            + cs.exp(-100 * (x - 1.5) ** 2)
-        )
-        nlp.init_solver(OPTS)
-        nlp2: Union[StackedMultistartNlp, ParallelMultistartNlp] = pickle.loads(
-            pickle.dumps(nlp)
-        )
-        self.assertEqual(nlp.name, nlp2.name)
-        if multinlp_cls is StackedMultistartNlp:
-            self.assertEqual(nlp._stacked_nlp.name, nlp2._stacked_nlp.name)
+        self.assertEqual(nlp.nh, expected_nh)
+        self.assertEqual(nlp.ng, expected_ng)
+        np.testing.assert_allclose(sol.vals["z"], 0.0, atol=1e-4, rtol=1e-4)
+        np.testing.assert_allclose(sol.vals["x"], expected_x, atol=1e-4, rtol=1e-4)
+        np.testing.assert_allclose(sol.vals["q"], 0.0, atol=1e-4, rtol=1e-4)
+        np.testing.assert_allclose(sol.vals["w"], 0.0, atol=1e-4, rtol=1e-4)
+        np.testing.assert_allclose(sol.vals["y"], expected_y, atol=1e-4, rtol=1e-4)
+        np.testing.assert_allclose(sol.vals["p"], 0.0, atol=1e-4, rtol=1e-4)
 
 
 if __name__ == "__main__":
